@@ -116,7 +116,10 @@ class UI:
         include_patterns.extend(r.pattern for r in torrent.include_regexs)
         if include_patterns:
             info('Include', include_patterns)
-        info('Files', self._fmt.files(torrent))
+        try:
+            info('Files', self._fmt.files(torrent))
+        except torf.PathError as e:
+            self.error(e, exit=False)
 
     def StatusReporter(self):
         if self._cfg['json'] or self._cfg['metainfo']:
@@ -154,6 +157,7 @@ class _FormatterBase:
     def terminate(self, torrent):
         pass
 
+
 class _HumanFormatter(_FormatterBase):
     def private(self, torrent):
         return 'yes' if torrent.private else 'no'
@@ -215,6 +219,7 @@ class _HumanFormatter(_FormatterBase):
         _term.echo('move_pos1')
         if newline:
             sys.stdout.write(f'{label}{LABEL_SEPARATOR}{value}\n')
+            _term.echo('ensure_line_below')
         else:
             sys.stdout.write(f'{label}{LABEL_SEPARATOR}{value}')
             _utils.flush(sys.stdout)
@@ -239,6 +244,7 @@ class _HumanFormatter(_FormatterBase):
             if answer is not None:
                 return answer
 
+
 class _MachineFormatter(_FormatterBase):
     def private(self, torrent):
         return 'yes' if torrent.private else 'no'
@@ -287,10 +293,13 @@ class _JSONFormatter(_MachineFormatter):
         return torrent.private
 
     def files(self, torrent):
-        return torrent.files
+        return (
+            {'Path': str(f), 'Size': f.size}
+            for f in torrent.files
+        )
 
     def info(self, key, value, newline=None):
-        # Join multiple values with a tab character
+        # Make sure we can JSON-encode all kinds of iterable
         if not isinstance(value, str) and isinstance(value, abc.Iterable):
             value = tuple(value)
         if key == 'Error':
@@ -328,17 +337,24 @@ class _MetainfoFormatter(_JSONFormatter):
 class _StatusReporterBase():
     def __init__(self, ui):
         self._ui = ui
+        self.reset()
+
+    def reset(self):
         self._start_time = time.time()
         self._progress = _utils.Average(samples=5)
         self._time_left = _utils.Average(samples=3)
         self._info = types.SimpleNamespace(
-            torrent=None, filepath=None,
-            pieces_done=0, pieces_total=0,
-            fraction_done=0, bytes_per_sec=0,
+            torrent=None,
+            filepath=None,
+            items_done=0,
+            items_total=0,
+            fraction_done=0,
+            throughput=0,
             time_left=datetime.timedelta(0),
             time_elapsed=datetime.timedelta(0),
             time_total=datetime.timedelta(0),
-            eta=datetime.datetime.now() + datetime.timedelta(300))
+            eta=datetime.datetime.now() + datetime.timedelta(300),
+        )
 
     def __enter__(self):
         return self
@@ -353,35 +369,62 @@ class _StatusReporterBase():
         pass
 
     def generate_callback(self, torrent, filepath, pieces_done, pieces_total):
-        self._update_progress_info(torrent, filepath, pieces_done, pieces_total)
-        self._ui.info('Progress', self._get_progress_string(self._info), newline=False)
+        self._update_progress_info_hashing(torrent, filepath, pieces_done, pieces_total)
+        progress_lines = self._get_hashing_progress_lines(self._info)
+        self._ui.info('Progress', progress_lines, newline=False)
+
+    def reuse_callback(self, torrent, torrent_filepath,
+                       torrent_files_done, torrent_files_total,
+                       is_match, exception):
+        if exception:
+            if isinstance(exception, torf.MetainfoError):
+                self._ui.info('Error', f'{torrent_filepath}: {self._format_error(exception, torrent)}')
+            else:
+                self._ui.info('Error', self._format_error(exception, torrent))
+
+        if is_match is True:
+            self._ui.info('Reused', torrent_filepath, newline=False)
+        elif is_match is None:
+            self._ui.info('Verifying', torrent_filepath, newline=False)
+        else:
+            self._update_progress_info_reuse(torrent, torrent_filepath,
+                                             torrent_files_done, torrent_files_total)
+            progress_lines = self._get_reuse_progress_lines(self._info)
+            self._ui.info('Reuse', progress_lines, newline=False)
 
     def verify_callback(self, torrent, filepath, pieces_done, pieces_total,
                         piece_index, piece_hash, exception):
         if exception:
             self._ui.info('Error', self._format_error(exception, torrent))
-        self._update_progress_info(torrent, filepath, pieces_done, pieces_total)
-        self._ui.info('Progress', self._get_progress_string(self._info), newline=False)
+        self._update_progress_info_hashing(torrent, filepath, pieces_done, pieces_total)
+        progress_lines = self._get_hashing_progress_lines(self._info)
+        self._ui.info('Progress', progress_lines, newline=False)
 
-    def _update_progress_info(self, torrent, filepath, pieces_done, pieces_total):
+    def _update_progress_info_common(self, torrent, filepath, items_done, items_total):
         info = self._info
         info.torrent = torrent
         info.filepath = filepath
-        info.pieces_done = pieces_done
-        info.pieces_total = pieces_total
-        info.fraction_done = pieces_done / pieces_total
-        progress = self._progress
+        info.items_done = items_done
+        info.items_total = items_total
+        if items_total > 0:
+            info.fraction_done = items_done / items_total
+        else:
+            info.fraction_done = 100.0
+
+    def _update_progress_info_hashing(self, torrent, filepath, pieces_done, pieces_total):
+        self._update_progress_info_common(torrent, filepath, pieces_done, pieces_total)
+        info = self._info
         if pieces_done < pieces_total:
-            progress.add(pieces_done)
+            self._progress.add(pieces_done)
             # Make sure we have enough samples to make estimates
-            if len(progress.values) >= 2:
+            if len(self._progress.values) >= 2:
                 info.time_elapsed = datetime.timedelta(seconds=round(time.time() - self._start_time))
-                time_diff = progress.times[-1] - progress.times[0]
-                pieces_diff = progress.values[-1] - progress.values[0]
+                time_diff = self._progress.times[-1] - self._progress.times[0]
+                pieces_diff = self._progress.values[-1] - self._progress.values[0]
                 bytes_diff = pieces_diff * torrent.piece_size
-                info.bytes_per_sec = bytes_diff / time_diff + 0.001  # Prevent ZeroDivisionError
+                info.throughput = bytes_diff / (time_diff + 0.001)  # Prevent ZeroDivisionError
                 bytes_left = (pieces_total - pieces_done) * torrent.piece_size
-                self._time_left.add(bytes_left / info.bytes_per_sec)
+                self._time_left.add(bytes_left / info.throughput)
                 info.time_left = datetime.timedelta(seconds=round(self._time_left.avg))
                 info.time_total = info.time_elapsed + info.time_left
                 info.eta = datetime.datetime.now() + info.time_left
@@ -389,13 +432,28 @@ class _StatusReporterBase():
             # The last piece was hashed
             info.time_elapsed = datetime.timedelta(seconds=round(time.time() - self._start_time))
             info.time_total = info.time_elapsed
-            info.bytes_per_sec = torrent.size / (info.time_total.total_seconds() + 0.001)  # Prevent ZeroDivisionError
+            info.throughput = torrent.size / (info.time_total.total_seconds() + 0.001)  # Prevent ZeroDivisionError
             info.time_left = datetime.timedelta(seconds=0)
             info.eta = datetime.datetime.now()
 
-    def _get_progress_string(self, info):
+    def _update_progress_info_reuse(self, torrent, filepath, files_done, files_total):
+        self._update_progress_info_common(torrent, filepath, files_done, files_total)
+        info = self._info
+        self._progress.add(files_done)
+        # Make sure we have enough samples to make estimates
+        if len(self._progress.values) >= 2:
+            info.time_elapsed = datetime.timedelta(seconds=round(time.time() - self._start_time))
+            time_diff = self._progress.times[-1] - self._progress.times[0]
+            files_diff = self._progress.values[-1] - self._progress.values[0]
+            info.throughput = files_diff / (time_diff + 0.001)  # Prevent ZeroDivisionError
+
+    def _get_hashing_progress_lines(self, info):
         return str(info)
 
+    def _get_reuse_progress_lines(self, info):
+        return str(info)
+
+
 class _HumanStatusReporter(_StatusReporterBase):
     def __enter__(self):
         _term.no_user_input.enable()
@@ -415,16 +473,18 @@ class _HumanStatusReporter(_StatusReporterBase):
     def __exit__(self, _, __, ___):
         _term.no_user_input.disable()
 
-    def _get_progress_string(self, info):
-        perc_str = f'{info.fraction_done * 100:5.2f} %'
-        bps_str = f'{_utils.bytes2string(info.bytes_per_sec, trailing_zeros=True)}/s'
-        if info.pieces_done < info.pieces_total:
-            term_width,_ = shutil.get_terminal_size()
-            term_width = min(term_width, 76)
-            # Available width minus label ("   Progress  ")
-            status_width = term_width - LABEL_WIDTH - len(LABEL_SEPARATOR)
+    def _get_status_width(self):
+        term_width, _ = shutil.get_terminal_size()
+        term_width = min(term_width, 76)
+        return term_width - LABEL_WIDTH - len(LABEL_SEPARATOR)
+
+    def _get_hashing_progress_lines(self, info):
+        percent_str = f'{info.fraction_done * 100:5.2f} %'
+        throughput_str = f'{_utils.bytes2string(info.throughput, trailing_zeros=True)}/s'
+        if info.items_done < info.items_total:
+            status_width = self._get_status_width()
             line1 = self._progress_line1(info.fraction_done, os.path.basename(info.filepath),
-                                         perc_str, bps_str, status_width)
+                                         percent_str, throughput_str, status_width)
             line2 = self._progress_line2(info, status_width)
             return ''.join((_term.save_cursor_pos,
                             _term.erase_to_eol,
@@ -433,18 +493,28 @@ class _HumanStatusReporter(_StatusReporterBase):
                             _term.restore_cursor_pos,
                             line1))
         else:
-            return STATUS_SEPARATOR.join((perc_str, f'{info.time_total} total', bps_str))
+            return STATUS_SEPARATOR.join((percent_str, f'{info.time_total} total', throughput_str))
 
-    def _progress_line1(self, fraction_done, filename, percent_str, bps_str, status_width):
-        progress_bar_width = (status_width - len('99.99 % ') - len(' 999.99 MiB/s') - 1)
+    def _get_reuse_progress_lines(self, info):
+        filename = os.path.basename(info.filepath)
+        term_width,_ = shutil.get_terminal_size()
+        term_width = min(term_width, 76)
+        # Available width minus label
+        status_width = term_width - LABEL_WIDTH - len(LABEL_SEPARATOR)
+        percent_str = f'{info.fraction_done * 100:5.2f} %'
+        throughput_str = f'{info.throughput:4.0f} files/s'
+        return self._progress_line1(info.fraction_done, filename,
+                                    percent_str, throughput_str, status_width)
+
+    def _progress_line1(self, fraction_done, filename, percent, suffix, status_width):
+        progress_bar_width = (status_width - len(percent) - len(suffix) - 1)
         if progress_bar_width >= 10:
-            progress_bar = self._progress_bar(filename, fraction_done,
-                                              width=progress_bar_width)
-            return ' '.join((percent_str, progress_bar, bps_str))
+            progress_bar = self._progress_bar(filename, fraction_done, progress_bar_width)
+            return ' '.join((percent, progress_bar, suffix))
         elif status_width >= 23:
-            return STATUS_SEPARATOR.join((percent_str, bps_str))
+            return STATUS_SEPARATOR.join((percent, suffix))
         else:
-            return percent_str
+            return percent
 
     def _progress_bar(self, text, fraction_done, width):
         text_width = width - 2
@@ -461,7 +531,7 @@ class _HumanStatusReporter(_StatusReporterBase):
                  'left'    : f'{info.time_left} left',
                  'total'   : f'{info.time_total} total',
                  'eta'     : f'ETA: {"{0:%H}:{0:%M}:{0:%S}".format(info.eta)}'}
-        priority = iter(('left', 'total', 'elapsed', 'eta'))
+        priority = ('left', 'total', 'elapsed', 'eta')
         order = ('elapsed', 'left', 'total', 'eta')
         parts = []
         priority = iter(priority)
@@ -498,16 +568,23 @@ class _HumanStatusReporter(_StatusReporterBase):
         else:
             return str(exception)
 
+
 class _MachineStatusReporter(_StatusReporterBase):
-    def _get_progress_string(self, info):
+    def _get_hashing_progress_lines(self, info):
         return '\t'.join((f'{info.fraction_done * 100:.3f}',
                           f'{round(info.time_elapsed.total_seconds())}',
                           f'{round(info.time_left.total_seconds())}',
                           f'{round(info.time_total.total_seconds())}',
                           f'{round(info.eta.timestamp())}',
-                          f'{round(info.bytes_per_sec)}',
+                          f'{round(info.throughput)}',
                           f'{info.filepath}'))
 
+    def _get_reuse_progress_lines(self, info):
+        return '\t'.join((f'{info.filepath}',
+                          f'{info.fraction_done * 100:.3f}',
+                          f'{info.items_done}',
+                          f'{info.items_total}'))
+
     def _format_error(self, exception, torrent):
         if isinstance(exception, torf.VerifyContentError) and len(exception.files) > 1:
             lines = [f'Corruption in piece {exception.piece_index + 1}, '
@@ -517,10 +594,16 @@ class _MachineStatusReporter(_StatusReporterBase):
         else:
             return str(exception)
 
+
 class _QuietStatusReporter(_MachineStatusReporter):
     def generate_callback(self, torrent, filepath, pieces_done, pieces_total):
         pass
 
+    def reuse_callback(self, torrent, torrent_filepath,
+                       torrent_files_done, torrent_files_total,
+                       is_match, exception):
+        pass
+
     def verify_callback(self, torrent, filepath, pieces_done, pieces_total,
                         piece_index, piece_hash, exception):
         if exception:
